/* amuxblks.c */
/* ############################################################### */
/* ############################################################### */
/* ###                                                         ### */
/* ###  IOMD AMUXBLKS                                          ### */
/* ###                                                         ### */
/* ###  Created 17/8/92: David Flynn (as SYSBLOCKS)            ### */
/* ###                                                         ### */
/* ###                                                         ### */
/* ############################################################### */
/* ############################################################### */
/* ###  Revision History:                                      ### */
/* ###   18/8/92: DWF : ADEC repartitioned : Sidle/C           ### */
/* ###   25/8/92: DWF : Import Ashley's AMUX & updates to ADEC ### */
/* ###   25/8/92: DWF : Import Ashley's AMUX & updates to ADEC ### */
/* ###   28/8/92: DWF : partion into separate blocks           ### */
/* ###   28/8/92:  AS : Change inccol to falling edge (of cas) ### */
/* ###   07/9/92:  AS : Add col address mask for video transfer### */
/* ###   25/11/92: AS : Removed column address masking         ### */
/*       13/01/93  WO : corrected mistakes in masking              */
/* ############################################################### */


#include <Handy.h>
#include "amuxblks.h"
#include "busarb.h"

/* ############################################################### */

typedef struct
{
  ModelCore     model;
  Signal        *dummy;

} *DelayModel;

static int get_delay (delay_name, model)
char *delay_name;
VoidStar model;
{
  Parameter    param;
  int          delay = 1000*NS;
  DelayModel   me = (DelayModel)model;

  param = simmodel_param(me->model.state,"delay");
  if (simmodel_type(param) == PT_INTEGER) return simmodel_int(param);
  if (simmodel_type(param) == PT_PARAMBLOCK) {
    param = simmodel_sub_param(param,delay_name);
    if (simmodel_type(param) == PT_INTEGER) return simmodel_int(param);
  }
  errhandler_f (SS_MODELS+DISASTER,"!!! %s delay not correctly specified\n",delay_name);
  return delay;
}


/* ############################################################### */
/* ############################################################### */
/* ### AMUX : IOMD Address Multiplexer                         ### */
/* ############################################################### */

#define AMUX_def \
"AMUX(\
  {IN}  ramc[],\
  {IN}  a[],\
  {IN}  da[],\
  {IN}  rclk,\
  {IN}  cas,\
  {IN}  dmagoing,\
  {IN}  colmux,\
  {IN}  fcolmux\
) => (\
  {OUT} count[],\
  {OUT} ra[],\
  {OUT} four,\
  {OUT} eight,\
  {OUT} maxseq\
)"

typedef struct
{
   ModelCore	model;
   Bundle	*ramc;
   Bundle	*a;
   Bundle	*da;
   Signal	*rclk;
   Signal	*cas;
   Signal	*dmagoing;
   Signal	*colmux;
   Signal	*fcolmux;
   Bundle	*count;
   Bundle	*ra;
   Signal	*four;
   Signal	*eight;
   Signal	*maxseq;
} *AMUX_model;

typedef struct
{
   int count, hi_col, lca, ca, row, four, eight, maxseq;
   int dt_cnt, dt_gt, dt_ltch, dt_mux;

} AMUX_state;

/*------------------------------------------------------------------------*/

static void AMUX_proc(modelInst)
VoidStar modelInst;
{
  AMUX_model me = (AMUX_model) modelInst;
  AMUX_state *s = (AMUX_state *)me->model.state;
       
  static int address;

  if (FALLING(rclk) && HI(colmux)) {
    SignalEvt(four, ((s->count & 0x3) == 0x2),  s->dt_ltch);
    SignalEvt(eight, ((s->count & 0x7) == 0x6), s->dt_ltch);
    SignalEvt(maxseq, ((s->count & 0xff) == 0xfe), s->dt_gt);
    ++s->count;
    BundleEvt(count, s->count, s->dt_cnt);
  }

  if LO(cas) s->ca = ((BINVAL(ramc) & 2) ? ( ((s->count>>1)&0xFF) | ((s->lca>>1)&0x300) )
                                         : ((s->count | s->lca) & 0x3ff));
  if LO(colmux) {
    if HI(dmagoing) address = BINVAL(da)<<2; else address = BINVAL(a)<<2;

    s->count = (address & 0x7fc) >> 2;
    BundleEvt(count, s->count, s->dt_cnt);
 
    s->hi_col = (((address & 0x1000000) >> 13) | ((address & 0x400000) >> 12));
    s->lca = ((address & 0x1c00) >> 2);

    s->row = ((address & 0x3ff000) >> 12);
    s->row = (s->row |
          ((address & 0x2000000) >> 14) |
          ((address & 0x800000) >> 13));
    if (BINVAL(ramc) & 1) s->row = s->row & 0xeff | ((address & 0x800) >> 3);
    if (BINVAL(ramc) & 4) s->row = s->row & 0xffe | ((address & 0x400000) >> 22);


    SignalEvt(four, ((s->count & 0x3) == 0x3), s->dt_ltch);
    SignalEvt(eight, ((s->count & 0x7) == 0x7), s->dt_ltch);
    SignalEvt(maxseq, ((s->count & 0xff) == 0xff), s->dt_gt);
  }

  if LO(fcolmux)
    BundleEvt(ra, s->row, s->dt_mux);
  else
    BundleEvt(ra, (s->hi_col | s->ca), s->dt_mux);

}

/*------------------------------------------------------------------------*/

static void AMUX_init(modelInst)
VoidStar modelInst;
{
  AMUX_state *s;
  AMUX_model me = (AMUX_model) modelInst;
  NEW(s, AMUX_state);
  me->model.proc = AMUX_proc;

  s->dt_cnt = DELAY("cnt"); 
  s->dt_gt  = DELAY("gt");
  s->dt_ltch= DELAY("ltch");
  s->dt_mux = DELAY("mux");
  me->model.state = (VoidStar ) s;

}


void AMUXblks_declare()
{
    simmodel_begin("iomd.amuxblks");
    simmodel_declare_s(AMUX_def,    AMUX_init);
    simmodel_end();
}

