/* ioctl.c */
/* ############################################################### */
/* ############################################################### */
/* ###                                                         ### */
/* ###  IOCTL block level model                                 ### */
/* ###                                                         ### */
/* ###  Created 27/8/92: Simon Segars                          ### */
/* ###                                                         ### */
/* ###                                                         ### */
/* ############################################################### */
/* ############################################################### */
/* ###  Revision History:                                      ### */
/* ###   27/8/92: SAS : Initial release                        ### */
/* ############################################################### */


#include <Handy.h>
#include "ioctl.h"
#include "busarb.h"

/* ############################################################### */

typedef struct
{
  ModelCore     model;
  Signal        *dummy;

} *DelayModel;

static int get_delay (delay_name, model)
char *delay_name;
VoidStar model;
{
  Parameter    param;
  int          delay = 1000*NS;
  DelayModel   me = (DelayModel)model;

  param = simmodel_param(me->model.state,"delay");
  if (simmodel_type(param) == PT_INTEGER) return simmodel_int(param);
  if (simmodel_type(param) == PT_PARAMBLOCK) {
    param = simmodel_sub_param(param,delay_name);
    if (simmodel_type(param) == PT_INTEGER) return simmodel_int(param);
  }
  errhandler_f (SS_MODELS+DISASTER,"!!! %s delay not correctly specified\n",delay_name);
  return delay;
}

#define IOCTL_def \
"IOCTL(da[1:0],dag_tc[3:0],data[7:0],dmach[2:0],drq[3:0],dsize[1:0],ioden[3:0],iodir[3:0],la[1:0],lah[15:0],\
Nfcas[3:0],snden[1:0],cas,cdrq,clk2,clk16,dmagoing,indma,into0,in2,into2,ioctlreg,iodec,io32dec,\
io_end,io_start,ioc_start,lwrite,memc_start,Ninit,Niogt,Niogti,Nior,Niorqi,Niow,Nscdack,Nspdack,pdrq,pio_free,prog,ramc,\
rclk,rdrq,ram_start,ready,readys,ref8m,reset,snd_end,snddir,snddrq,sndrw,snd_start,swdone,was8ck,wasW,free,freed,bus,biodma) \
=> (Ndack[3:0],Nfrd[3:0],fwt_lat[3:0],fwt_s[1:0],ioch[1:0],end_io,frd_oe,frd_bh,iodrq,lrnw,\
Nblw,Nccs,Ncdack,Neasis,Nenbd,Niorq,Nmemc,Nms,Npboe,Nrbe,Nsccs,Nsio,Nsndcs,Nwbe,sound,snd0drq,snd1drq,tc,dir,ioend,\
bytes,ioc[1:0],pio[2:0],spd[1:0],dma[3:0],dmasm[1:0],iochan[3:0],scount[1:0],drqs[3:0])"

typedef struct {
    ModelCore   model;
    Bundle      *da, *dag_tc, *data, *dmach, *drq, *dsize, *ioden, *iodir, *la, *lah, *Nfcas, *snden;
    Signal      *cas, *cdrq, *clk2, *clk16, *dmagoing, *indma, *into0, *in2, *into2, *ioctlreg, *iodec;
    Signal      *io32dec, *io_end, *io_start, *ioc_start, *lwrite, *memc_start, *Ninit,  *Niogt, *Niogti;
    Signal      *Nior, *Niorqi, *Niow, *Nscdack, *Nspdack, *pdrq, *pio_free;
    Signal      *prog,*ramc,*rclk,*rdrq,*ram_start,*ready,*readys,*ref8m,*reset,*snd_end, *snddir;
    Signal      *snddrq, *sndrw, *snd_start, *swdone, *was8ck, *wasW, *free, *freed, *bus, *biodma;
    Bundle      *Ndack, *Nfrd, *fwt_lat, *fwt_s, *ioch;
    Signal      *end_io, *frd_oe, *frd_bh, *iodrq, *lrnw, *Nblw;
    Signal      *Nccs, *Ncdack, *Neasis, *Nenbd, *Niorq, *Nmemc, *Nms, *Npboe, *Nrbe, *Nsccs, *Nsio, *Nsndcs;
    Signal      *Nwbe, *sound, *snd0drq, *snd1drq, *tc, *dir, *ioend, *bytes;
    Bundle      *ioc,*pio,*spd,*dma,*dmasm,*iochan,*scount,*drqs;
    } *IOCTL_model;

typedef struct {
    int dt_gt;
    int dack0, dma0;
    int dmaext, dmatcr, iotcr, ectcr;
    int iodrql, rrobin;
    int freel, fin, finish, drqs, drqll, dir, ldir, dack_state, speed, dma_state;
    int iodma, iodmau, cs, written, serviced;
    int pio_state, progio;
    int dal, done, rdone, end, io_end, io_endl, memcs, memcl, memcll;
    int sio, ioc_state;
    int wasWRL0, ram_cyc, drql, io_startu, iostartl;
    int dag_tcl, ioc_nrwl, iock;
    int sndch, sread, snd_state, snd_count, cdrql, cdrqll, cdrqs, pdrql, pdrqll, pdrqs;
    int rbel_eas, rbe_state, bytes, sndip, sndipu, sdone, snd_startu, snd_startl;
    int end_of_io, frd_state;
    int snd_countl, sndrwl, sndrws;
    int readyl, frd_dma, frd_prog;
    } IOCTL_state;

/*--------------------------------------------------------------------------*/

static void IOCTL_proc (modelInst)
VoidStar modelInst;
{
    IOCTL_model me = (IOCTL_model) modelInst;
    IOCTL_state *s = (IOCTL_state *)me->model.state;
    int fprog=0, nrw=0, ioc_spd=0, Add=0, drqch=0;

/* ------------The IOCTL Timing registers ------------------------*/

  if (HI(prog) && HI(ioctlreg)) {
      switch (BINVAL(la)) {
         case 0 : if HI(lwrite)
                     s->dmatcr = BINVAL(data);
                    else
                     BundleEvt(data, s->dmatcr, s->dt_gt);
                  break;
         case 1 : if HI(lwrite)
                     s->iotcr = BINVAL(data);
                    else
                     BundleEvt(data, s->iotcr, s->dt_gt);
                  break;
         case 2 : if HI(lwrite)
                     s->ectcr = BINVAL(data);
                    else
                     BundleEvt(data, s->ectcr, s->dt_gt);
                  break;
         case 3 : if HI(lwrite)
                     s->dmaext = BINVAL(data)&0xF;
                    else
                     BundleEvt(data, s->dmaext, s->dt_gt);
                  break;
         }
       }

  /* Free up the bus after ioctl writes onto it */
  SignalEvt(bus, HI(ioctlreg) && HI(prog) && LO(lwrite), s->dt_gt);
  if FALLING(bus)
     BundleZZZ(data, s->dt_gt);


/*-----------------Round-robin arbitration of drq's------------------------*/
  if HI(reset) {
     BundleEvt(Ndack, 0xF, s->dt_gt);
     BundleEvt(ioch, 0, s->dt_gt);
     SignalEvt(Nccs, TRUE, s->dt_gt);
     SignalEvt(Ncdack, TRUE, s->dt_gt);
     SignalEvt(Nsccs, TRUE, s->dt_gt);
     SignalEvt(Neasis, TRUE, s->dt_gt);
     SignalEvt(Niow, TRUE, s->dt_gt);
     SignalEvt(Nior, TRUE, s->dt_gt);
     SignalEvt(Nsndcs, TRUE, s->dt_gt);
     }

  if RISING(free)
     s->dir = FALSE;
     
/* general purpose IO channels only arbitrated if no sound requests */

  if ((s->drqs!=0) && !(s->cdrqs || s->pdrqs) && HI(Niorqi) && s->freel && RISING(rclk)) {
     /* Mask out the disabled channels into a local variable */
     drqch = (0xF & (BINVAL(ioden) & s->drqs));
     if (drqch != 0) {
      s->iodmau = TRUE;
      switch (s->rrobin) {
       case 0 :  if ((drqch&0x2)==0x2)
                    s->rrobin = 1;
                 else if ((drqch&0x4)==0x4)
                    s->rrobin = 2;
                 else if ((drqch&0x8)==0x8)
                    s->rrobin = 3;
                 else
                    s->rrobin = 0; /* Just for completeness */
                 break;

       case 1 :  if ((drqch&0x4)==0x4)
                    s->rrobin = 2;
                 else if ((drqch&0x8)==0x8)
                    s->rrobin = 3;
                 else if ((drqch&0x1)==0x1)
                    s->rrobin = 0;
                 else 
                    s->rrobin = 1;
                 break;

       case 2 :  if ((drqch&0x8)==0x8)
                    s->rrobin = 3;
                 else if ((drqch&0x1)==0x1)
                    s->rrobin = 0;
                 else if ((drqch&0x2)==0x2)
                    s->rrobin = 1;
                 else
                    s->rrobin = 2;
                 break;

       case 3 :  if ((drqch&0x1)==0x1)
                    s->rrobin = 0;
                 else if ((drqch&0x2)==0x2)
                    s->rrobin = 1;
                 else if ((drqch&0x4)==0x4)
                    s->rrobin = 2;
                 else
                    s->rrobin = 3;
                 break;
       }
       BundleEvt(ioch, s->rrobin, s->dt_gt);
       /* s->speed = (s->dmatcr>>2*s->rrobin)&0x3;
       BundleEvt(spd, s->speed, s->dt_gt); */
       s->dir = (BINVAL(iodir) >> s->rrobin)&1;
      }
   }

/* The two sound channels */
   if RISING(snddrq) {
      s->sndch = (s->cdrqs==TRUE);
      s->dir = (s->cdrqs || (!s->cdrqs && s->pdrqs && HI(snddir)));
      }

   if (HI(reset) || RISING(free))
      s->sndipu = FALSE;
   if (HI(free) && HI(snddrq))
      s->sndipu = TRUE;

   if (HI(reset) || RISING(free))
      s->sndip = FALSE;
   if LO(rclk)
      s->sndip = s->sndipu;

SignalEvt(snd_end, s->sdone, s->dt_gt);

   /* SignalEvt(dir, s->dir, s->dt_gt); Put this last thing */

/* The block global control signals, all SRFFs, just for fun */

/* SRFF on free */
  if ((HI(free) && (HI(snddrq) || (drqch!=0))) || HI(pio_free) /*HI(into2)*/)
      SignalEvt(free, FALSE, s->dt_gt);
  if (HI(reset) || HI(into0) || ((s->dir==WRITE) && HI(swdone)) ||
        ((s->dir==READ) && HI(freed)) || ((s->iodma) && HI(freed)))
      SignalEvt(free, TRUE, s->dt_gt);

/* Latched version */
  if LO(rclk)
     s->freel = HI(free);

  if (RISING(freed) || HI(reset))
     s->iodma = FALSE;

/* SRFF on io_start */

   s->end = (HI(reset) || (s->dack_state==STATE_WRL1) || ((s->dack_state==STATE_DL0)
               && (s->speed==SPEED_D)));

   if (((s->dma_state==2) && (s->iodma) && (s->dir==WRITE)) || ((s->iodma) && (s->dir==READ) && (s->done==FALSE)) 
       || (HI(into2) && ((s->progio==8) || (s->progio<6))))
      s->io_startu = TRUE; /* Set */
   if (s->end)
      s->io_startu = FALSE; /* Reset */

/* Now sync to clk16 */
  if (RISING(clk16) || (s->end))
     s->iostartl = s->io_startu;
  if (LO(clk16) || (s->end))
     SignalEvt(io_start, s->iostartl, s->dt_gt);

/* SRFF on ram_start */
  if (HI(rclk) && ((s->iodma) || (s->sndip))) {
     if (((s->dir==WRITE) && (s->dack_state==STATE_DH) && (s->dma_state==0) && LO(io_start) && (s->rdone==FALSE) &&
        (s->done==FALSE)) || ((s->dir==READ) && ((s->sdone) || (s->ram_cyc))))
     SignalEvt(ram_start, TRUE, s->dt_gt);
     }
  if (HI(reset) || (s->dma_state==1))
     SignalEvt(ram_start, FALSE, s->dt_gt);

/* Now sync snd_start to 16MHz  Note : evaluation order important here */
  if (RISING(clk16) || (s->snd_state==8))
     s->snd_startl = s->snd_startu;
  if (LO(clk16) || (s->snd_state==8))
     SignalEvt(snd_start, s->snd_startl, s->dt_gt);

/* SRFF on snd_start */
  if ((((s->dir==READ) && !(s->sread)) || ((s->dma_state==2) && (s->dir==WRITE))) && (s->sndip))
     s->snd_startu = TRUE;
  if (HI(reset) || (s->snd_state==8))
     s->snd_startu = FALSE;

  if FALLING(snd_start)
     s->sread = TRUE;
  if HI(free)
     s->sread = FALSE;

/* Another SRFF on memc_start */

  if (HI(into2) && (s->progio==6))
     SignalEvt(memc_start, TRUE, s->dt_gt);
  if (HI(reset) || s->io_end)
     SignalEvt(memc_start, FALSE, s->dt_gt);

/* Now sync to 16MHz */
  if LO(clk16)
     s->memcs = s->memcl;
  if RISING(clk16)
     s->memcl = HI(memc_start);

/*-------------------------- DMA RAM Cycle State Machine ----------------------------------------------------*/

  if FALLING(rclk)
     s->dma0 = (s->dma_state==0);

  if HI(reset)
     s->dma_state = 0;
      
  if FALLING(rclk) {
     switch (s->dma_state) {
            case 0 : if (((s->sndip) || (s->iodma)) && HI(ram_start))
                        s->dma_state = 1;
                     break;
            case 1 : if (HI(dmagoing) && ((s->sndip && ((BINVAL(dmach)==SND1DMA) || (BINVAL(dmach)==SND0DMA))) || 
                        (s->iodma && (BINVAL(dmach)==IODMA)))) {
                        if ((s->dir==READ) && (s->dack_state==STATE_DH))
                           s->dma_state = 0;
                          else
                           s->dma_state = 2;
                        }
                     break;
            case 2 : if ((s->dir==WRITE) || (s->dack_state==STATE_DH));
                        s->dma_state = 0;
                     break;
            }
     }

  BundleEvt(dmasm, s->dma_state, s->dt_gt);
  SignalEvt(rdrq, (s->dma_state==1), s->dt_gt);

  SignalEvt(snd0drq, (HI(rdrq) && (s->sndch==TRUE) && (s->sndip)), s->dt_gt);
  SignalEvt(snd1drq, (HI(rdrq) && (s->sndch==FALSE) && (s->sndip)), s->dt_gt);
  SignalEvt(iodrq, (HI(rdrq) && (s->iodma)), s->dt_gt);

/*---------------------- 16MHz DMA State Machine --------------------------*/

/* In here for evaluation order sake */
  if (RISING(clk16) && (((s->speed==SPEED_A) && (s->dack_state==STATE_WRH)) ||
     ((s->speed!=SPEED_A) && HI(readys) && (s->dack_state==STATE_WRL0))))
     s->fin=!s->iodma;
  if (HI(rclk) && (s->pio_state==0))
     s->fin=FALSE;
     
 /* READY Synchronisation */

  if RISING(clk16)
     s->readyl = HI(ready);

  if LO(clk16)
     SignalEvt(readys, s->readyl, s->dt_gt);

  /* Evaluate this before terms to kick off the state machine */
  if RISING(wasW)
     s->wasWRL0 = TRUE;
  if (HI(free) || HI(ram_start))
     s->wasWRL0 = FALSE;

 /* The terms to kick off the ram cyc state machine */
   if (HI(reset) || HI(ram_start))
      s->ram_cyc = 0;
   if (LO(clk16) && (s->iodma) && HI(dir))
      if (((s->dack_state==STATE_DH) && HI(io_start) && (s->speed==SPEED_D)) ||
                    ((s->dack_state==STATE_DL0) && (s->speed==SPEED_C)) ||
                    ((s->wasWRL0) && ((s->speed==SPEED_A) || (s->speed==SPEED_B))))
          s->ram_cyc = TRUE;

  if RISING(clk16)
     SignalEvt(wasW, (HI(dir) && (s->dack_state==STATE_WRL0) && !(HI(iodec) && LO(readys))), s->dt_gt);

SignalEvt(ramc, s->ram_cyc, s->dt_gt);

 /* The state machine itself */

  if HI(reset)
     s->dack_state = STATE_DH;

  if (RISING(clk16)) {
     switch (s->dack_state) {
       case STATE_DH : if HI(io_start)
                          s->dack_state = STATE_DL0;
                       break;
       case STATE_DL0 : if (s->speed==SPEED_D) {
                           s->dack_state = STATE_WRL0;
                           }
                          else if (s->speed==SPEED_C)
                           s->dack_state = STATE_WRL1;
                          else
                           s->dack_state = STATE_DL1;
                        break;
       case STATE_DL1 : if (s->speed==SPEED_A)
                           s->dack_state = STATE_WRL3;
                          else
                           s->dack_state = STATE_WRL2;
                       break;
       case STATE_WRL0 : if (HI(iodec) && LO(readys))
                            s->dack_state = STATE_WRL0;
                           else if ((s->speed==SPEED_C) || (s->speed==SPEED_D))
                            s->dack_state = STATE_DH;
                          else {
                            if (s->speed==SPEED_A)
                              s->dack_state = STATE_WRH;
                             else if HI(iodec)
                              s->dack_state = STATE_DH;
                             else
                              s->dack_state = STATE_SI;
                             }
                         break;
       case STATE_WRL3 : s->dack_state = STATE_WRL2;
                         break;
       case STATE_WRL2 : s->dack_state = STATE_WRL1;
                         break;
       case STATE_WRL1 : s->dack_state = STATE_WRL0;
                         break;
       case STATE_WRH : if HI(iodec)
                           s->dack_state = STATE_DH;
                         else
                           s->dack_state = STATE_SI;
                        break;
       case STATE_SI : s->dack_state = STATE_DH;
                       break;
       }
     }

/* Primary State machine outputs */
BundleEvt(dma, s->dack_state, s->dt_gt);

  if LO(clk16)
     s->dack0 = (s->dack_state==0);

  if ((s->dir==READ) && FALLING(rdrq))
     s->serviced = TRUE;
  if HI(free)
     s->serviced = FALSE;

  if FALLING(io_start)
     s->done=TRUE;
  if HI(free)
     s->done=FALSE;

  if (((s->dir==WRITE) && ((s->iodma) || (s->progio!=0))) && (s->done))
     s->written = TRUE;
  if HI(free)
     s->written = FALSE;

  if (((s->serviced) || (s->written)) && (s->dma0) && (s->dack0))
     SignalEvt(freed, TRUE, s->dt_gt);
  if HI(free)
     SignalEvt(freed, FALSE, s->dt_gt);

/* De-mux to make chip selects, dacks etc */

   if FALLING(clk16)
      s->cs = ((s->dack_state&0x8)!=0x8);

   switch (s->progio&0xF) {
      case 5 : SignalEvt(Neasis, s->cs, s->dt_gt);
              break;
      case 1 : SignalEvt(Nccs, s->cs, s->dt_gt);
              break;
      case 2 : SignalEvt(Ncdack, s->cs, s->dt_gt);
              break;
      case 3 : SignalEvt(Ncdack, s->cs, s->dt_gt);
               /* Assert TC here also */
              break;
      case 4 : SignalEvt(Nsccs, s->cs, s->dt_gt);
              break;
      case 0 : if (s->sndip==FALSE)
                  BundleEvt(Ndack, ~(!s->cs<<s->rrobin), s->dt_gt);
              break;
      case 8 : SignalEvt(Nsndcs, s->cs, s->dt_gt);
              break;
      }

/* Synchronise the incoming DRQ's - must be evaluated AFTER state machine */

  if RISING(rclk) {
     s->drql = BINVAL(drq);
     s->cdrql = (HI(cdrq) && ((BINVAL(snden)&0x1)==0x1));
     s->pdrql = (HI(pdrq) && ((BINVAL(snden)&0x2)==0x2));
     SignalEvt(snddrq, (HI(free) && HI(Niorqi) && (s->pdrqs || s->cdrqs)), s->dt_gt);
     }
  if LO(rclk) {
     s->drqll = s->drql;
     s->cdrqll = s->cdrql;
     s->pdrqll = s->pdrql;
     }
  s->drqs = (s->drqll & BINVAL(drq)); 
  s->cdrqs = (s->cdrqll && HI(cdrq));
  s->pdrqs = (s->pdrqll && HI(pdrq));
  
BundleEvt(drqs, s->drqs, s->dt_gt);


/*-------------------Programmed IO ---------------------------------------*/

  /* if (RISING(clk16) && (((s->speed==SPEED_A) && (s->dack_state==STATE_WRH)) ||
     ((s->speed!=SPEED_A) && HI(readys) && (s->dack_state==STATE_WRL0))))
     s->fin=!s->iodma;
  if (HI(rclk) && (s->pio_state==0))
     s->fin=FALSE; */
     

  if (HI(rclk) && (s->iodma==FALSE) && HI(iodec))
     s->finish = s->fin;
  if (HI(rclk) && (s->pio_state==0))
     s->finish = FALSE;

  if (s->pio_state==0)
     s->finish = FALSE;

  if FALLING(rclk)
     SignalEvt(into0, ((s->pio_state==3) && LO(Niorqi)), s->dt_gt);
  if HI(free)
     SignalEvt(into0, FALSE, s->dt_gt);

  if FALLING(rclk)
     SignalEvt(pio_free, (HI(free) && LO(Niorqi) && (s->pio_state<2)), s->dt_gt);

  if HI(reset) {
     s->pio_state = 0;
     s->progio = 0;
     }
    else if FALLING(rclk) {
     switch (s->pio_state) {
        case 0 : SignalEvt(Niogti, TRUE, s->dt_gt);
                 if LO(Niorqi) { 
                    if HI(free)
                       s->pio_state = 2;
                      else
                       s->pio_state = 1;
                 }
                 break;
        case 1 : if HI(Niorqi)
                    s->pio_state = 0;
                 else if HI(free)
                    s->pio_state = 2;
                 break;
        case 2 : if (s->end_of_io==TRUE)
                    s->pio_state = 3;
                 break;
        case 3 : if LO(Niorqi)
                    s->pio_state = 0;
                 break;
        }
    }

  BundleEvt(pio, s->pio_state, s->dt_gt);

  if FALLING(rclk)
  s->end_of_io = ((((s->progio<6) || (s->progio==8)) && (s->progio!=0) && (s->finish) && LO(io_start)) ||
                     (LO(memc_start) && (s->progio==6)) ||
                     (LO(ioc_start) && (s->progio==7)));

  SignalEvt(end_io, s->end_of_io, s->dt_gt);

  SignalEvt(Niogti, (s->pio_state!=3), s->dt_gt);

  if HI(rclk)
     SignalEvt(in2, (s->pio_state==2), s->dt_gt);

  if RISING(in2)
     SignalEvt(into2, TRUE, s->dt_gt);
  if (HI(reset) || (LO(rclk) && HI(in2)))
     SignalEvt(into2, FALSE, s->dt_gt);


/*----------8MHz 'old' style state machine that runs at 16MHz really------*/

/*  if ((s->io_end) || LO(clk16))
     SignalEvt(Niorq, (!s->memcll || s->io_endl), s->dt_gt); */

  if FALLING(ref8m)
     s->io_endl = s->io_end;

/* D-type with reset */
   if (s->io_endl)
      SignalEvt(Niorq, TRUE, s->dt_gt);
    else if FALLING(clk16)
      SignalEvt(Niorq, !s->memcll, s->dt_gt);

  if RISING(ref8m)
     s->memcll = s->memcs;

  if ((s->memcs) && RISING(ref8m))
     s->io_end = (LO(Niorq) && LO(Niogt));
  if (HI(reset) || HI(free))
     s->io_end = FALSE;

  SignalEvt(io_end, s->io_end, s->dt_gt);

/* ----------- IOC type 'simple' podule interface ------------------------*/

/* Get the cycle timing from the address */

  ioc_spd = ((BINVAL(lah)&0x0180)>>7)&0x3;
  BundleEvt(ioc, ioc_spd, s->dt_gt);

  /* SRFF on ioc_start */

  if (HI(into2) && (s->sio==0) && (s->progio==7))
     SignalEvt(ioc_start, TRUE, s->dt_gt);
  if (HI(reset) || (HI(clk16) && (s->sio==3)))
    SignalEvt(ioc_start, FALSE, s->dt_gt);

/* The 8MHz state machine */

  if HI(reset)
       s->ioc_state = 0;
    else {
     if RISING(ref8m) {
        switch (s->ioc_state) {
           case 0 : if HI(ioc_start) {
                       if (ioc_spd!=3)
                          s->ioc_state = 1;
                       else if (s->iock==1)
                          s->ioc_state = 7;
                    }
                    break;
           case 1 : if (ioc_spd==0)
                       s->ioc_state = 4;
                     else
                       s->ioc_state = 7;
                    break;
           case 3 : s->ioc_state = 0;
                    break;
           case 4 : s->ioc_state = 7;
                    break;
           case 6 : if ((ioc_spd&0x2)==0x2)
                       s->ioc_state = 0;
                      else
                       s->ioc_state = 3;
                    break;
           case 7 : s->ioc_state = 6;
                    break;
           }
        }
     }

/* Retime Nsio to 16MHz, from small state machine */

    if HI(reset)
       s->sio = 0;
      else if FALLING(clk16) {
         switch (s->sio) {
            case 0 : if (s->ioc_state!=0) 
                        s->sio = 2;
                     break;
            case 2 : if (s->ioc_state==0)
                        s->sio = 3;
                     break;
            case 3 : s->sio = 0;
                     break;
            }
      }

/* Retime ior/w in for sync case */
   if FALLING(ref8m)
      s->ioc_nrwl = ((s->ioc_state&0x2)==0x2);

  SignalEvt(Nsio, s->sio==0, s->dt_gt);
 
/* Make the 2MHz clock - must be evaluated after 8M state machine */

  if LO(Ninit)
     s->iock = 0;
   else if FALLING(ref8m) {
     s->iock = (s->iock+1)&0x3;
     SignalEvt(clk2, (s->iock&0x2)==0x2, 2*s->dt_gt);
     }

/*-------- And just when you thought everything was done, here comes the sound block ----------*/

   if HI(reset)
      s->snd_state = 0;
    else if RISING(clk16) {
         switch (s->snd_state) {
            case 0 : if HI(snd_start)
                        if (s->dir==READ)
                           s->snd_state = 1;
                          else
                           s->snd_state = 0xC;
                     break;
            case 0xC : s->snd_state = 0xD;
                       break;
            case 0xD : s->snd_state = 1;
                       break;
            case 1 : s->snd_state = 2;
                     break;
            case 2 : s->snd_state = 3;
                     break;
            case 3 : s->snd_state = 4;
                     break;
            case 4 : s->snd_state = 8;
                     break;
            case 5 : s->snd_state = 2;
                     break;
            case 8 : if (s->bytes==TRUE)
                        s->snd_state = 0;
                       else
                        s->snd_state = 5;
                     break;
            }
         }

  if (LO(clk16) && (s->bytes==TRUE) && ((s->snd_state&0x8)==8))
     s->sdone = (s->dir==READ);
  if (HI(reset) || (s->dma_state==1))
     s->sdone = FALSE;

  if LO(clk16)
     SignalEvt(was8ck, ((s->bytes==TRUE) && (s->snd_state==8)), s->dt_gt);

  if FALLING(was8ck)
     SignalEvt(swdone, TRUE, s->dt_gt);
  if HI(free)
     SignalEvt(swdone, FALSE, s->dt_gt);

  if FALLING(clk16) {
     if (s->sndch==FALSE) {
        SignalEvt(Nspdack, ((s->snd_state==0) || (s->snd_state==0xC) || (s->snd_state==0xD)), s->dt_gt);
        SignalEvt(Nscdack, TRUE, s->dt_gt);
        }
       else {
        SignalEvt(Nscdack, ((s->snd_state==0) || (s->snd_state==0xC) || (s->snd_state==0xD)), s->dt_gt);
        SignalEvt(Nspdack, TRUE, s->dt_gt);
        }
    }

  if (FALLING(rdrq) && (s->dir==WRITE) && (s->sndip))
     s->rdone = TRUE;
  if HI(free)
     s->rdone = FALSE;

  SignalEvt(sndrw, (s->snd_state&0x2)==0x2, s->dt_gt);
  if FALLING(clk16)
     s->sndrwl = HI(sndrw);

  s->sndrws = (HI(sndrw) || s->sndrwl);

  /* Two bit counter, re-loaded from ioctl[7:6] when snd dacks are both high */

   if (s->snd_state==0)
      s->snd_count = 0;
    else if FALLING(sndrw)
       s->snd_count = (s->snd_count+1)&0x3;

BundleEvt(scount, s->snd_count, s->dt_gt);

   s->bytes = (s->snd_count==((s->iotcr>>6)&0x3));
SignalEvt(bytes, s->bytes, s->dt_gt);

   if RISING(sndrw)
      s->snd_countl = s->snd_count; 


/*---------------------------------------------------------------------------------------------*/

/* Nior/w etc, plus funnel oe control */

   nrw = !(((s->dack_state&0x4)==0x4) || (((s->ioc_state&0x2)==0x2) && (ioc_spd!=3)) ||
            ((ioc_spd==3) && s->ioc_nrwl) || HI(sndrw));

   if (s->dir==WRITE) {
      SignalEvt(Niow, nrw, s->dt_gt);
      }
     else {
      SignalEvt(Nior, nrw, s->dt_gt);
      }

  /* State machine for frd_oe */

  if HI(reset)
     s->frd_state=0;

  fprog = ((s->dir==READ) && LO(Niogti) && LO(Niorqi));


  if RISING(rclk) {
     switch (s->frd_state) {
       case 0 : if ((s->dir==READ) && ((HI(dmagoing) && 
                    ((BINVAL(dmach)==IODMA) || (s->sndip && (BINVAL(dmach)==SND0DMA) || (BINVAL(dmach)==SND1DMA))))))
                   s->frd_state=1;
                break;
       case 1 : s->frd_state=0;
                break;
       }
    }

 if FALLING(rclk)
    s->frd_dma = ((s->frd_state&1)==1);

  SignalEvt(frd_oe, ((HI(iodec) && fprog) || s->frd_dma), s->dt_gt);

/* -------- Programmed IO Address decoder --------------------------------*/

  if (((s->pio_state&2)==2) && HI(iodec)) {
     SignalEvt(dir, (s->dir = LO(lwrite)), s->dt_gt);
     Add = (BINVAL(lah) & 0xFFFF);
     if (HI(io32dec) && ((Add & 0x8000)==0x8000)) {
         s->speed = ((s->ectcr >> ((Add & 0x7000) >> 12))&0x1) ? SPEED_C : SPEED_A;
         BundleEvt(spd, s->speed, s->dt_gt);
         s->progio = 5;
         }
     else if (HI(iodec) && ((Add & 0xF000)==0x3000)) {
        if ((Add & 0x0FF0)==0x0010) {
           if ((Add & 0x000F)<0x2)
              s->progio = 1; /* Combo CS */
             else
              s->progio = 2; /* Combo DACK */
           }
        else if ((Add & 0x0FF0)==0x0020) {
           if ((Add & 0x000F)==0xA)
              s->progio = 3; /* Combo DACK + TC */
            else if ((Add & 0x000F)==0xB)
              s->progio = 4; /* SCSI CS */
            else if ((Add & 0x000F)>=0xC)
              s->progio = 8; /* Sound codec */
            else if ((Add & 0x000F) <0xA)
              s->progio = 2; /* The rest of Combo DACK */
           }
        else if (((Add & 0x00F0)==0x0000) || ((Add & 0x00F0)==0x0030))
           s->progio = 6; /* Internal MEMC podules */
        else if ((Add & 0x0FF0)>=0x0210)
           s->progio = 7; /* Simple Peripherals */
        }
     BundleEvt(iochan, s->progio, s->dt_gt);
     }

    if (HI(into0) || HI(reset))
       BundleEvt(iochan, (s->progio=0), s->dt_gt);

    SignalEvt(Nmemc, s->progio!=6, s->dt_gt);

/* Speed Multiplexor */

   if (s->progio==8)
     s->speed = (s->iotcr>>4)&0x3;
    else if (s->progio==4)
     s->speed = s->iotcr & 0x3;
    else if ((s->progio<4) && (s->progio!=0))
           s->speed = (s->iotcr>>2)&0x3;

  if (s->progio==0)
     s->speed = (s->dmatcr>>2*s->rrobin)&0x3;

BundleEvt(spd, s->speed, s->dt_gt);

/*------------------------TC Control-------------------------*/ 

  if (HI(rdrq) && s->iodma) /* Transparent latch with reset */
     s->dag_tcl = BINVAL(dag_tc);
  if HI(reset)
     s->dag_tcl = 0;

  SignalEvt(tc, (!s->cs && ((HI(iodec) && (s->progio==3)) || (s->iodma & (s->dag_tcl>>s->rrobin)))),s->dt_gt);

/*-----------------------Nms Control----------------------------------*/

  SignalEvt(Nms, (s->progio!=6), s->dt_gt);

/*-------------------Funnel and Buffer Control Signals--------------------------------*/
/* Reset for now */
  if HI(reset) {
     BundleEvt(fwt_s, 0, s->dt_gt);
/*     SignalEvt(frd_oe, FALSE, s->dt_gt); */
     /* SignalEvt(frd_bh, FALSE, s->dt_gt);*/
     }

  SignalEvt(Nenbd, s->dir, s->dt_gt);

  SignalEvt(sound, s->sndip, s->dt_gt);

  SignalEvt(Nwbe, (s->dir && (s->iodma || (s->progio==5))), s->dt_gt);

  if LO(dmagoing)
    s->ldir = s->dir;

/* Changed ldir to dir in below */
  if FALLING(rclk)
     s->rbe_state = ((s->dir==READ) && (BINVAL(dsize)==0) && HI(dmagoing) && (BINVAL(dmach)==IODMA) && !s->rbe_state);

  if FALLING(rclk)
     s->rbel_eas = ((s->progio==5) && (s->dir==READ) && LO(Niogti) && LO(Niorqi));

  SignalEvt(Nrbe, !(s->rbel_eas || ((s->progio==5) && (s->dir==READ) && LO(Niogti) && LO(Niorqi)) 
                    || s->rbe_state), s->dt_gt);

  SignalEvt(Nblw, ((s->dir==WRITE) && (HI(cas) && HI(dmagoing) && ((s->dma_state==2) && s->iodma)) || ((s->progio==5) && HI(into2))), 2*s->dt_gt);

 /* Funnel write latch control */

  /* Make stretched signals to say if you're doing a sound or a io dma */
  if RISING(rclk)
     SignalEvt(indma, (HI(dmagoing) && (( (((BINVAL(dmach)==SND0DMA) && ((BINVAL(snden)&0x1)==0x1)) || (BINVAL(dmach)==SND1DMA))) ||
                          (s->iodma && (BINVAL(dmach)==IODMA)))), s->dt_gt);

  if (s->dir==WRITE) {
     if HI(indma)
        BundleEvt(fwt_lat, ~BINVAL(Nfcas), 2*s->dt_gt); 
       else
        BundleEvt(fwt_lat, (0xF * HI(into2)), 2*s->dt_gt); /* Programmed IO */
    }
   else
      BundleEvt(fwt_lat, 0x0, s->dt_gt);

  if ((s->sndip==FALSE) && (s->progio!=6))
     BundleEvt(Nfrd, ~(LO(Nior) * 0xF), s->dt_gt);

  if ((s->dir==READ) && (s->sndip)) {
     if (s->snd_count==0)
        BundleEvt(Nfrd, ~(0xF & (0xF * HI(sndrw))), s->dt_gt);
     if (s->snd_count==1)
        BundleEvt(Nfrd, ~(0xA  & (0xF * HI(sndrw))), s->dt_gt);
     if (s->snd_count==2)
        BundleEvt(Nfrd, ~(0xC  & (0xF * HI(sndrw))), s->dt_gt);
     if (s->snd_count==3)
        BundleEvt(Nfrd, ~(0x8  & (0xF * HI(sndrw))), s->dt_gt);
     }

  /* for the case of MEMC peripherals, frd is handled in the funnel */

  GATHER(da);

  if ((s->progio==7) || (s->progio==6))
     BundleEvt(fwt_s, 2, s->dt_gt);

  if ((s->progio==8) || ((s->progio<=5) && (s->progio>0)))
     BundleEvt(fwt_s, 0, s->dt_gt);

  if HI(reset)
     s->dal=0;
  else if ((s->iodma || s->sndip) && HI(rdrq)) {
     if (VAL(da)==UNDEFINED)
        s->dal=8;
       else
        s->dal = (BINVAL(da)&3);
     }

  if (s->iodma || ((s->sndip==0) && (s->progio==0))) {
     if (s->dal==8)
        BundleXXX(fwt_s, 3*s->dt_gt);
       else
        BundleEvt(fwt_s, s->dal, 3*s->dt_gt);
      }

  if ((s->sndip) && LO(rdrq)) {
     if (s->dal==8)
        BundleXXX(fwt_s, s->dt_gt);
       else
        BundleEvt(fwt_s, (0xF * (s->sndrws)) & (s->snd_countl | s->dal), s->dt_gt);
     }

  if ((s->progio==0) && (s->sndip==0) && (s->iodma==0)) {
     if (s->dal==8)
        BundleXXX(fwt_s, 3*s->dt_gt);
       else
        BundleEvt(fwt_s, s->dal, 3*s->dt_gt);
      }

  if (HI(dmagoing) && ((BINVAL(dmach)==SND0DMA) || (BINVAL(dmach)==SND1DMA)))
     SignalEvt(frd_bh, TRUE, s->dt_gt);
   else if (HI(dmagoing) && (BINVAL(dmach)==IODMA)) {
     if (VAL(da)==UNDEFINED)
        SignalEvt(frd_bh, UNDEFINED, s->dt_gt);
       else
        SignalEvt(frd_bh, (BINVAL(da)&0x1)==0, s->dt_gt);
      }
   else
     SignalEvt(frd_bh, TRUE, s->dt_gt);

 /*-----Npboe buffer--------*/
 /* Three cases : iodma controlled by dmaext register, IOC controlled by la[18], and MEMC podules */
 /* The buffer is always enabled during writes. */

  SignalEvt(Npboe, !((s->iodma && ((s->dmaext>>s->rrobin)&1)) || (s->dir==WRITE) ||
                    ((s->progio==7) && ((BINVAL(lah)&0x40)==0x40)) ||
                    (s->progio==6) || (s->progio==5)), s->dt_gt);

  /* lrnw, used to be in with the MEMC stuff now here due to causality probs in verification */

  SignalEvt(lrnw, s->dir, s->dt_gt); /* Can it really be this simple ? */

/* moved this past FALLING(clk16) as rclk will probably be later ! */

  if FALLING(rclk)
     s->iodma = s->iodmau;
  if (HI(reset) || HI(freed))
     s->iodmau = FALSE;

  /* More causality probs */
  SignalEvt(dir, s->dir, s->dt_gt); 

  SignalEvt(biodma, s->iodma, 1);

/*-------------- All done --------------------------*/
}

static void IOCTL_init(modelInst)
VoidStar modelInst;
{
  IOCTL_state *s;
  IOCTL_model me = (IOCTL_model) modelInst;
  NEW(s, IOCTL_state);
  me->model.proc = IOCTL_proc;
  s->dt_gt = DELAY("gt");
  s->rrobin = 0;
  s->dal=8;

  me->model.state = (VoidStar ) s;

}

void IOCTLC_declare()
{
    simmodel_begin("iomd.ioctlc");
    simmodel_declare_s(IOCTL_def, IOCTL_init);
    simmodel_end();
}


/*--------------------------------------------------------------------------*/

